@@ -138,6 +138,17 @@ static void entryDraw(newtComponent co) {
     }
 
     chptr = en->buf + en->firstChar;
+
+    if (en->flags & NEWT_FLAG_PASSWORD) {
+	char *tmpptr, *p;
+
+	tmpptr = alloca(strlen(chptr+2));
+	strcpy(tmpptr, chptr);
+	for (p = tmpptr; *p; p++)
+	    *p = '*';
+	chptr = tmpptr;
+    }			
+
     len = strlen(chptr);
 
     if (len <= co->width) {
@@ -190,7 +201,7 @@ static struct eventResult entryEvent(newtComponent co,
     if (ev.when == EV_NORMAL) {
 	switch (ev.event) {
 	case EV_FOCUS:
-	    /*SLtt_set_cursor_visibility(0);*/
+	    newtCursorOn();
 	    if (en->flags & NEWT_FLAG_HIDDEN)
 		newtGotorc(co->top, co->left);
 	    else
@@ -200,7 +211,7 @@ static struct eventResult entryEvent(newtComponent co,
 	    break;
 
 	case EV_UNFOCUS:
-	    /*SLtt_set_cursor_visibility(1);*/
+	    newtCursorOff();
 	    newtGotorc(0, 0);
 	    er.result = ER_SWALLOWED;
 	    if (co->callback)
