@@ -4,6 +4,7 @@
 #include <stdarg.h>
 #include <stdlib.h>
 #include <sys/select.h>
+#include <sys/time.h>
 
 #ifdef USE_GPM
 #include <ctype.h>
@@ -395,6 +396,10 @@ struct form {
     int numFds;
     struct fdInfo * fds;
     int maxFd;
+    int timer;    /* in milliseconds */
+    struct timeval lastTimeout;
+    void * helpTag;
+    newtCallback helpCb;
 };
 
 static void gotoComponent(struct form * form, int newComp);
@@ -402,6 +407,9 @@ static struct eventResult formEvent(newtComponent co, struct event ev);
 static struct eventResult sendEvent(newtComponent comp, struct event ev);
 static void formPlace(newtComponent co, int left, int top);
 
+/* Global, ick */
+static newtCallback helpCallback;
+
 /* this isn't static as grid.c tests against it to find forms */
 struct componentOps formOps = {
     newtDrawForm,
@@ -422,7 +430,7 @@ static inline int componentFits(newtComponent co, int compNum) {
     return 1;
 }
 
-newtComponent newtForm(newtComponent vertBar, const char * help, int flags) {
+newtComponent newtForm(newtComponent vertBar, void * help, int flags) {
     newtComponent co;
     struct form * form;
 
@@ -455,6 +463,8 @@ newtComponent newtForm(newtComponent vertBar, const char * help, int flags) {
     form->background = COLORSET_WINDOW;
     form->hotKeys = malloc(sizeof(int));
     form->numHotKeys = 0;
+    form->timer = 0;
+    form->lastTimeout.tv_sec = form->lastTimeout.tv_usec = 0;
     if (!(form->flags & NEWT_FLAG_NOF12)) {
 	newtFormAddHotKey(co, NEWT_KEY_F12);
     }
@@ -464,6 +474,9 @@ newtComponent newtForm(newtComponent vertBar, const char * help, int flags) {
     else
 	form->vertBar = NULL;
 
+    form->helpTag = help;
+    form->helpCb = helpCallback;
+
     return co;
 }
 
@@ -494,6 +507,14 @@ void newtFormSetCurrent(newtComponent co, newtComponent subco) {
     gotoComponent(form, new);
 }
 
+void newtFormSetTimer(newtComponent co, int millisecs) {
+    struct form * form = co->data;
+
+    form->timer = millisecs;
+    form->lastTimeout.tv_usec = 0;
+    form->lastTimeout.tv_sec = 0;
+}
+
 void newtFormSetHeight(newtComponent co, int height) {
     struct form * form = co->data;
 
@@ -864,6 +885,7 @@ void newtFormRun(newtComponent co, struct newtExitStruct * es) {
     int key, i, max;
     int done = 0;
     fd_set readSet, writeSet;
+    struct timeval nextTimeout, now, timeout;
 #ifdef USE_GPM
     int x, y;
     Gpm_Connect conn;
@@ -909,9 +931,49 @@ void newtFormRun(newtComponent co, struct newtExitStruct * es) {
 		FD_SET(form->fds[i].fd, &writeSet);
 	}
 
-	i = select(max + 1, &readSet, &writeSet, NULL, NULL);
+	if (form->timer) {
+	    /* Calculate when we next need to return with a timeout. Do
+	       this inside the loop in case a callback resets the timer. */
+	    if (!form->lastTimeout.tv_sec && !form->lastTimeout.tv_usec)
+		gettimeofday(&form->lastTimeout, NULL);
+
+	    nextTimeout.tv_sec = form->lastTimeout.tv_sec + 
+		    (form->timer / 1000);
+	    nextTimeout.tv_usec = form->lastTimeout.tv_usec + 
+				    (form->timer % 1000) * 1000;
+
+	    gettimeofday(&now, 0);
+
+	    if (now.tv_sec > nextTimeout.tv_sec) {
+		timeout.tv_sec = timeout.tv_usec = 0;
+	    } else if (now.tv_sec == nextTimeout.tv_sec) {
+		timeout.tv_sec = 0;
+		if (now.tv_usec > nextTimeout.tv_usec)
+		    timeout.tv_usec = 0;
+		else
+		    timeout.tv_usec = nextTimeout.tv_usec - now.tv_usec;
+	    } else if (now.tv_sec < nextTimeout.tv_sec) {
+		timeout.tv_sec = nextTimeout.tv_sec - now.tv_sec;
+		if (now.tv_usec > nextTimeout.tv_usec)
+		    timeout.tv_sec--,
+		    timeout.tv_usec = nextTimeout.tv_usec + 1000000 -
+					now.tv_usec;
+		else 
+		    timeout.tv_usec = nextTimeout.tv_usec - now.tv_usec;
+	    }
+	} else {
+	    timeout.tv_sec = timeout.tv_usec = 0;
+	}
+
+	i = select(max + 1, &readSet, &writeSet, NULL, 
+			form->timer ? &timeout : NULL);
 	if (i < 0) continue;	/* ?? What should we do here? */
 
+	if (i == 0) {
+	    done = 1;
+	    es->reason = NEWT_EXIT_TIMER;
+	    gettimeofday(&form->lastTimeout, NULL);
+	} else
 #ifdef USE_GPM
 	if (gpm_fd > 0 && FD_ISSET(gpm_fd, &readSet)) {
 	    Gpm_GetEvent(&event);
@@ -956,6 +1018,9 @@ void newtFormRun(newtComponent co, struct newtExitStruct * es) {
 		    }
 		}
 
+		if (key == NEWT_KEY_F1 && form->helpTag && form->helpCb)
+		    form->helpCb(co, form->helpTag);
+
 		if (!done) {
 		    ev.event = EV_KEYPRESS;
 		    ev.u.key = key;
@@ -1039,3 +1104,7 @@ void newtFormWatchFd(newtComponent co, int fd, int fdFlags) {
     form->fds[form->numFds++].flags = fdFlags;
     if (form->maxFd < fd) form->maxFd = fd;
 }
+
+void newtSetHelpCallback(newtCallback cb) {
+    helpCallback = cb;
+}
