@@ -11,6 +11,13 @@ typedef struct snackWidget_s snackWidget;
 typedef struct snackGrid_s snackGrid;
 typedef struct snackForm_s snackForm;
 
+struct callbackStruct {
+    PyObject * cb, * data;
+};
+
+static struct callbackStruct suspend;
+static struct callbackStruct helpCallback;
+
 static void emptyDestructor(PyObject * s);
 
 static snackWidget * buttonWidget(PyObject * s, PyObject * args);
@@ -19,6 +26,9 @@ static snackWidget * checkboxWidget(PyObject * s, PyObject * args);
 static PyObject * choiceWindow(PyObject * s, PyObject * args);
 static snackWidget * entryWidget(PyObject * s, PyObject * args);
 static PyObject * drawRootText(PyObject * s, PyObject * args);
+static PyObject * doResume(PyObject * s, PyObject * args);
+static PyObject * doSuspend(PyObject * s, PyObject * args);
+static PyObject * doSuspend(PyObject * s, PyObject * args);
 static snackForm * formCreate(PyObject * s, PyObject * args);
 static snackGrid * gridCreate(PyObject * s, PyObject * args);
 static PyObject * gridWrappedWindow(PyObject * s, PyObject * args);
@@ -33,10 +43,15 @@ static PyObject * popWindow(PyObject * s, PyObject * args);
 static PyObject * pushHelpLine(PyObject * s, PyObject * args);
 static snackWidget * radioButtonWidget(PyObject * s, PyObject * args);
 static PyObject * refreshScreen(PyObject * s, PyObject * args);
+static PyObject * scaleWidget(PyObject * s, PyObject * args);
+static PyObject * scaleSet(snackWidget * s, PyObject * args);
 static PyObject * screenSize(PyObject * s, PyObject * args);
+static PyObject * setSuspendCallback(PyObject * s, PyObject * args);
+static PyObject * setHelpCallback(PyObject * s, PyObject * args);
 static PyObject * reflowText(PyObject * s, PyObject * args);
 static snackWidget * textWidget(PyObject * s, PyObject * args);
 static PyObject * ternaryWindow(PyObject * s, PyObject * args);
+static snackWidget * checkboxTreeWidget(PyObject * s, PyObject * args);
 
 static PyMethodDef snackModuleMethods[] = {
     { "button", (PyCFunction) buttonWidget, METH_VARARGS, NULL },
@@ -49,6 +64,7 @@ static PyMethodDef snackModuleMethods[] = {
     { "form", (PyCFunction) formCreate, METH_VARARGS, NULL },
     { "grid", (PyCFunction) gridCreate, METH_VARARGS, NULL },
     { "gridwrappedwindow", gridWrappedWindow, METH_VARARGS, NULL },
+    { "helpcallback", setHelpCallback, METH_VARARGS, NULL },
     { "init", initScreen, METH_VARARGS, NULL },
     { "label", (PyCFunction) labelWidget, METH_VARARGS, NULL },
     { "listbox", (PyCFunction) listboxWidget, METH_VARARGS, NULL },
@@ -60,9 +76,14 @@ static PyMethodDef snackModuleMethods[] = {
     { "radiobutton", (PyCFunction) radioButtonWidget, METH_VARARGS, NULL },
     { "reflow", (PyCFunction) reflowText, METH_VARARGS, NULL },
     { "refresh", refreshScreen, METH_VARARGS, NULL },
+    { "resume", doResume, METH_VARARGS, NULL },
+    { "scale", scaleWidget, METH_VARARGS, NULL },
     { "size", screenSize, METH_VARARGS, NULL },
+    { "suspend", doSuspend, METH_VARARGS, NULL },
+    { "suspendcallback", setSuspendCallback, METH_VARARGS, NULL },
     { "ternary", ternaryWindow, METH_VARARGS, NULL },
     { "textbox", (PyCFunction) textWidget, METH_VARARGS, NULL },
+    { "checkboxtree", (PyCFunction) checkboxTreeWidget, METH_VARARGS, NULL },
     { NULL }
 } ;
 
@@ -105,13 +126,19 @@ struct snackForm_s {
 
 static PyObject * formGetAttr(PyObject * s, char * name);
 static PyObject * formAdd(snackForm * s, PyObject * args);
+static PyObject * formDraw(snackForm * s, PyObject * args);
 static PyObject * formRun(snackForm * s, PyObject * args);
 static PyObject * formHotKey(snackForm * s, PyObject * args);
+static PyObject * formSetCurrent(snackForm * form, PyObject * args);
+static PyObject * formSetTimer(snackForm * form, PyObject * args);
 
 static PyMethodDef formMethods[] = {
     { "add", (PyCFunction) formAdd, METH_VARARGS, NULL },
+    { "draw", (PyCFunction) formDraw, METH_VARARGS, NULL },
     { "run", (PyCFunction) formRun, METH_VARARGS, NULL },
     { "addhotkey", (PyCFunction) formHotKey, METH_VARARGS, NULL },
+    { "setcurrent", (PyCFunction) formSetCurrent, METH_VARARGS, NULL },
+    { "settimer", (PyCFunction) formSetTimer, METH_VARARGS, NULL },
     { NULL }
 };
 
@@ -138,25 +165,58 @@ struct snackWidget_s {
     char achar;
     void * apointer;
     int anint;
+    struct callbackStruct scs;
 } ;
 
 static PyObject * widgetAddCallback(snackWidget * s, PyObject * args);
 static PyObject * widgetGetAttr(PyObject * s, char * name);
+static void widgetDestructor(PyObject * s);
 static PyObject * widgetEntrySetValue(snackWidget * s, PyObject * args);
+static PyObject * widgetLabelText(snackWidget * s, PyObject * args);
 static PyObject * widgetListboxSetW(snackWidget * s, PyObject * args);
 static PyObject * widgetListboxAdd(snackWidget * s, PyObject * args);
 static PyObject * widgetListboxIns(snackWidget * s, PyObject * args);
 static PyObject * widgetListboxDel(snackWidget * s, PyObject * args);
 static PyObject * widgetListboxGet(snackWidget * s, PyObject * args);
+static PyObject * widgetListboxSet(snackWidget * s, PyObject * args);
+static PyObject * widgetTextboxText(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeAddItem(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeGetSel(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeGetCur(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeSetEntry(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeSetEntryValue(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxTreeGetEntryValue(snackWidget * s, PyObject * args);
+static PyObject * widgetEntrySetFlags(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxSetFlags(snackWidget * s, PyObject * args);
+static PyObject * widgetCheckboxSetValue(snackWidget * s, PyObject * args);
 
 static PyMethodDef widgetMethods[] = {
     { "setCallback", (PyCFunction) widgetAddCallback, METH_VARARGS, NULL },
+    { "labelText", (PyCFunction) widgetLabelText, METH_VARARGS, NULL },
+    { "textboxText", (PyCFunction) widgetTextboxText, METH_VARARGS, NULL },
     { "entrySetValue", (PyCFunction) widgetEntrySetValue, METH_VARARGS, NULL },
     { "listboxAddItem", (PyCFunction) widgetListboxAdd, METH_VARARGS, NULL },
     { "listboxInsertItem", (PyCFunction) widgetListboxIns, METH_VARARGS, NULL },
     { "listboxGetCurrent", (PyCFunction) widgetListboxGet, METH_VARARGS, NULL },
+    { "listboxSetCurrent", (PyCFunction) widgetListboxSet, METH_VARARGS, NULL },
     { "listboxSetWidth", (PyCFunction) widgetListboxSetW, METH_VARARGS, NULL },
     { "listboxDeleteItem", (PyCFunction) widgetListboxDel, METH_VARARGS, NULL },
+    { "scaleSet", (PyCFunction) scaleSet, METH_VARARGS, NULL },
+    { "checkboxtreeAddItem", (PyCFunction) widgetCheckboxTreeAddItem,
+      METH_VARARGS, NULL },
+    { "checkboxtreeGetCurrent", (PyCFunction) widgetCheckboxTreeGetCur,
+      METH_VARARGS, NULL },
+    { "checkboxtreeGetEntryValue", (PyCFunction) widgetCheckboxTreeGetEntryValue,
+      METH_VARARGS, NULL },
+    { "checkboxtreeSetEntry", (PyCFunction) widgetCheckboxTreeSetEntry,
+      METH_VARARGS, NULL },
+    { "checkboxtreeSetEntryValue", (PyCFunction) widgetCheckboxTreeSetEntryValue,
+      METH_VARARGS, NULL },
+    { "checkboxtreeGetSelection", (PyCFunction) widgetCheckboxTreeGetSel,
+      METH_VARARGS, NULL },
+    { "entrySetFlags", (PyCFunction) widgetEntrySetFlags, METH_VARARGS, NULL },
+    { "checkboxSetFlags", (PyCFunction) widgetCheckboxSetFlags, METH_VARARGS, NULL },
+    { "checkboxSetValue", (PyCFunction) widgetCheckboxSetValue, METH_VARARGS, NULL },
     { NULL }
 };
 
@@ -166,7 +226,7 @@ static PyTypeObject snackWidgetType = {
         "snackwidget",                  /* tp_name */
         sizeof(snackWidget),            /* tp_size */
         0,                              /* tp_itemsize */
-        emptyDestructor,      		/* tp_dealloc */
+        widgetDestructor,      		/* tp_dealloc */
         0,                              /* tp_print */
         widgetGetAttr,  		/* tp_getattr */
         0,                              /* tp_setattr */
@@ -177,11 +237,21 @@ static PyTypeObject snackWidgetType = {
         0,                		/* tp_as_mapping */
 };
 
-static void callbackEntry(newtComponent co, void * arg) {
-    PyEval_CallObject(arg, arg);
+static snackWidget * snackWidgetNew (void) {
+    snackWidget * widget;
+     
+    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+
+    widget->scs.cb = NULL;
+    widget->scs.data = NULL;
+
+    return widget;
 }
 
 static PyObject * initScreen(PyObject * s, PyObject * args) {
+    suspend.cb = NULL;
+    suspend.data = NULL;
+    
     newtInit();
     newtCls();
 
@@ -190,6 +260,9 @@ static PyObject * initScreen(PyObject * s, PyObject * args) {
 }
 
 static PyObject * finishScreen(PyObject * s, PyObject * args) {
+    Py_XDECREF (suspend.cb);
+    Py_XDECREF (suspend.data);
+    
     newtFinished();
     Py_INCREF(Py_None);
     return Py_None;
@@ -201,6 +274,29 @@ static PyObject * refreshScreen(PyObject * s, PyObject * args) {
     return Py_None;
 }
 
+static PyObject * scaleWidget(PyObject * s, PyObject * args) {
+    snackWidget * widget;
+    int width, fullAmount;
+
+    if (!PyArg_ParseTuple(args, "ii", &width, &fullAmount)) return NULL;
+
+    widget = snackWidgetNew ();
+    widget->co = newtScale(-1, -1, width, fullAmount);
+
+    return (PyObject *) widget;
+}
+
+static PyObject * scaleSet(snackWidget * s, PyObject * args) {
+    int amount;
+
+    if (!PyArg_ParseTuple(args, "i", &amount)) return NULL;
+
+    newtScaleSet(s->co, amount);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * screenSize(PyObject * s, PyObject * args) {
     int width, height;
 
@@ -212,6 +308,88 @@ static PyObject * screenSize(PyObject * s, PyObject * args) {
     return Py_BuildValue("(ii)", width, height);
 }
 
+static void helpCallbackMarshall(newtComponent co, void * data) {
+    PyObject * args, * result;
+
+    args = Py_BuildValue("(O)", data);
+    result = PyEval_CallObject(helpCallback.cb, args);
+    Py_DECREF (args);
+    Py_XDECREF(result);
+
+    return;
+}
+
+static void suspendCallbackMarshall(void * data) {
+    struct callbackStruct * scs = data;
+    PyObject * args, * result;
+
+    if (scs->data) {
+	args = Py_BuildValue("(O)", scs->data);
+	result = PyEval_CallObject(scs->cb, args);
+	Py_DECREF (args);
+    } else
+	result = PyEval_CallObject(scs->cb, NULL);
+    
+    if (!result) {
+	PyErr_Print();
+	PyErr_Clear();
+    }
+
+    Py_XDECREF(result);
+
+    return;
+}
+
+static void callbackMarshall(newtComponent co, void * data) {
+    struct callbackStruct * scs = data;
+    PyObject * args, * result;
+
+    if (scs->data) {
+	args = Py_BuildValue("(O)", scs->data);
+	result = PyEval_CallObject(scs->cb, args);
+	Py_DECREF (args);
+    } else
+	result = PyEval_CallObject(scs->cb, NULL);
+
+    if (!result) {
+	PyErr_Print();
+	PyErr_Clear();
+    }
+    
+    Py_XDECREF(result);
+
+    return;
+}
+
+static PyObject * setSuspendCallback(PyObject * s, PyObject * args) {
+    if (!PyArg_ParseTuple(args, "O|O", &suspend.cb, &suspend.data))
+	return NULL;
+
+    Py_INCREF (suspend.cb);
+    Py_XINCREF (suspend.data);    
+    
+    newtSetSuspendCallback(suspendCallbackMarshall, &suspend);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * setHelpCallback(PyObject * s, PyObject * args) {
+    if (!PyArg_ParseTuple(args, "O", &helpCallback.cb))
+	return NULL;
+
+    /*if (helpCallback.cb) {
+	Py_DECREF (helpCallback.cb);
+    }*/
+
+    Py_INCREF (helpCallback.cb);
+
+    newtSetHelpCallback(helpCallbackMarshall);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * drawRootText(PyObject * s, PyObject * args) {
     int left, top;
     char * text;
@@ -225,6 +403,20 @@ static PyObject * drawRootText(PyObject * s, PyObject * args) {
     return Py_None;
 }
 
+static PyObject * doSuspend(PyObject * s, PyObject * args) {
+    newtSuspend();
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * doResume(PyObject * s, PyObject * args) {
+    newtResume();
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * popHelpLine(PyObject * s, PyObject * args) {
     newtPopHelpLine();
     Py_INCREF(Py_None);
@@ -279,11 +471,16 @@ static PyObject * centeredWindow(PyObject * s, PyObject * args) {
 static PyObject * gridWrappedWindow(PyObject * s, PyObject * args) {
     snackGrid * grid;
     char * title;
+    int x = -1, y = -1;
 
-    if (!PyArg_ParseTuple(args, "O!s", &snackGridType, &grid, &title))
+    if (!PyArg_ParseTuple(args, "O!s|ii", &snackGridType, &grid, &title, 
+			  &x, &y))
 	return NULL;
 
-    newtGridWrappedWindow(grid->grid, title);
+    if (y == -1)
+	newtGridWrappedWindow(grid->grid, title);
+    else
+	newtGridWrappedWindowAt(grid->grid, title, x, y);
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -355,7 +552,7 @@ static snackWidget * buttonWidget(PyObject * s, PyObject * args) {
 
     if (!PyArg_ParseTuple(args, "s", &label)) return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtButton(-1, -1, label);
 
     return widget;
@@ -367,12 +564,34 @@ static snackWidget * labelWidget(PyObject * s, PyObject * args) {
 
     if (!PyArg_ParseTuple(args, "s", &label)) return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtLabel(-1, -1, label);
 
     return widget;
 }
 
+static PyObject * widgetLabelText(snackWidget * s, PyObject * args) {
+    char * label;
+
+    if (!PyArg_ParseTuple(args, "s", &label)) return NULL;
+
+    newtLabelSetText(s->co, label);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * widgetTextboxText(snackWidget * s, PyObject * args) {
+    char * text;
+
+    if (!PyArg_ParseTuple(args, "s", &text)) return NULL;
+
+    newtTextboxSetText(s->co, text);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static snackWidget * listboxWidget(PyObject * s, PyObject * args) {
     snackWidget * widget;
     int height;
@@ -381,9 +600,9 @@ static snackWidget * listboxWidget(PyObject * s, PyObject * args) {
     if (!PyArg_ParseTuple(args, "i|ii", &height, &doScroll, &returnExit))
 	return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtListbox(-1, -1, height,
-				(doScroll ? 0 : NEWT_FLAG_NOSCROLL) |
+				(doScroll ? NEWT_FLAG_SCROLL : 0) |
 				(returnExit ? NEWT_FLAG_RETURNEXIT : 0));
     widget->anint = 1;
     
@@ -394,14 +613,16 @@ static snackWidget * textWidget(PyObject * s, PyObject * args) {
     char * text;
     int width, height;
     int scrollBar = 0;
+    int wrap = 0;
     snackWidget * widget;
     
-    if (!PyArg_ParseTuple(args, "iis|i", &width, &height, &text, &scrollBar))
+    if (!PyArg_ParseTuple(args, "iis|ii", &width, &height, &text, &scrollBar, &wrap))
 	return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtTextbox(-1, -1, width, height,
-				scrollBar ? NEWT_FLAG_SCROLL : 0);
+				(scrollBar ? NEWT_FLAG_SCROLL : 0) |
+ 			        (wrap ? NEWT_FLAG_WRAP : 0));
     newtTextboxSetText(widget->co, text);
     
     return widget;
@@ -415,7 +636,7 @@ static snackWidget * radioButtonWidget(PyObject * s, PyObject * args) {
     if (!PyArg_ParseTuple(args, "sOi", &text, &group, &isOn)) 
 		return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
 
     if ((PyObject *) group == Py_None)
 	widget->co = newtRadiobutton(-1, -1, text, isOn, NULL);
@@ -432,36 +653,65 @@ static snackWidget * checkboxWidget(PyObject * s, PyObject * args) {
 
     if (!PyArg_ParseTuple(args, "si", &text, &isOn)) return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtCheckbox(-1, -1, text, isOn ? '*' : ' ', NULL, 
 				&widget->achar);
 
     return widget;
 }
 
+static PyObject * widgetCheckboxSetFlags(snackWidget * s, PyObject * args) {
+    int flag, sense;
+
+    if (!PyArg_ParseTuple(args, "ii", &flag, &sense)) return NULL;
+
+    newtCheckboxSetFlags(s->co, flag, sense);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * widgetCheckboxSetValue(snackWidget * s, PyObject * args) {
+    char *value;
+
+    if (!PyArg_ParseTuple(args, "s", &value)) return NULL;
+
+    newtCheckboxSetValue(s->co, *value);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static snackWidget * entryWidget(PyObject * s, PyObject * args) {
     snackWidget * widget;
     int width;
     char * initial;
-    int isHidden, isScrolled, returnExit;
+    int isHidden, isScrolled, returnExit, isPassword;
 
-    if (!PyArg_ParseTuple(args, "isiii", &width, &initial,
-			  &isHidden, &isScrolled, &returnExit)) return NULL;
+    if (!PyArg_ParseTuple(args, "isiiii", &width, &initial,
+			  &isHidden, &isPassword, &isScrolled, &returnExit)) return NULL;
 
-    widget = PyObject_NEW(snackWidget, &snackWidgetType);
+    widget = snackWidgetNew ();
     widget->co = newtEntry(-1, -1, initial, width, (char **) &widget->apointer, 
 			   (isHidden ? NEWT_FLAG_HIDDEN : 0) |
+			   (isPassword ? NEWT_FLAG_PASSWORD : 0) |
 			   (returnExit ? NEWT_FLAG_RETURNEXIT : 0) |
-			   (!isScrolled ? NEWT_FLAG_NOSCROLL : 0));
+			   (isScrolled ? NEWT_FLAG_SCROLL : 0));
 
     return widget;
 }
 
 static snackForm * formCreate(PyObject * s, PyObject * args) {
     snackForm * form;
+    PyObject * help = Py_None;
+
+    if (!PyArg_ParseTuple(args, "|O", &help)) return NULL;
+
+    if (help == Py_None)
+	help = NULL;
 
     form = PyObject_NEW(snackForm, &snackFormType);
-    form->fo = newtForm(NULL, NULL, 0);
+    form->fo = newtForm(NULL, help, 0);
 
     return form;
 }
@@ -524,6 +774,13 @@ static PyObject * formGetAttr(PyObject * s, char * name) {
     return Py_FindMethod(formMethods, s, name);
 }
 
+static PyObject * formDraw(snackForm * s, PyObject * args) {
+    newtDrawForm(s->fo);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * formAdd(snackForm * s, PyObject * args) {
     snackWidget * w;
     int size = PyTuple_Size(args), i;
@@ -550,6 +807,8 @@ static PyObject * formRun(snackForm * s, PyObject * args) {
 
     if (result.reason == NEWT_EXIT_HOTKEY)
 	return Py_BuildValue("(si)", "hotkey", result.u.key);
+    else if (result.reason == NEWT_EXIT_TIMER)
+	return Py_BuildValue("(si)", "timer", 0);
     else
 	return Py_BuildValue("(si)", "widget", result.u.co);
 }
@@ -566,6 +825,30 @@ static PyObject * formHotKey(snackForm * s, PyObject * args) {
     return Py_None;
 }
 
+static PyObject * formSetTimer(snackForm * form, PyObject * args) {
+    int millisecs;
+
+    if (!PyArg_ParseTuple(args, "i", &millisecs))
+	return NULL;
+
+    newtFormSetTimer(form->fo, millisecs);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * formSetCurrent(snackForm * form, PyObject * args) {
+    snackWidget * w;
+
+    if (!PyArg_ParseTuple(args, "O", &w))
+	return NULL;
+
+    newtFormSetCurrent(form->fo, w->co);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * widgetGetAttr(PyObject * s, char * name) {
     snackWidget * w = (snackWidget *) s;
 
@@ -582,13 +865,26 @@ static PyObject * widgetGetAttr(PyObject * s, char * name) {
     return Py_FindMethod(widgetMethods, s, name);
 }
 
-static PyObject * widgetAddCallback(snackWidget * s, PyObject * args) {
-    PyObject * object;
+static void widgetDestructor(PyObject * o) {
+    snackWidget * s = (snackWidget *) o;
+    
+    Py_XDECREF (s->scs.cb);
+    Py_XDECREF (s->scs.data);
 
-    if (!PyArg_ParseTuple(args, "O", &object))
+    PyMem_DEL(o);
+}
+
+static PyObject * widgetAddCallback(snackWidget * s, PyObject * args) {
+    s->scs.cb = NULL;
+    s->scs.data = NULL;
+    
+    if (!PyArg_ParseTuple(args, "O|O", &s->scs.cb, &s->scs.data))
 	return NULL;
 
-    newtComponentAddCallback(s->co, callbackEntry, object);
+    Py_INCREF (s->scs.cb);
+    Py_XINCREF (s->scs.data);
+    
+    newtComponentAddCallback(s->co, callbackMarshall, &s->scs);
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -606,6 +902,18 @@ static PyObject * widgetEntrySetValue(snackWidget * s, PyObject * args) {
     return Py_None;
 }
 
+static PyObject * widgetEntrySetFlags(snackWidget * s, PyObject * args) {
+    int flag, sense;
+
+    if (!PyArg_ParseTuple(args, "ii", &flag, &sense)) return NULL;
+
+    newtEntrySetFlags(s->co, flag, sense);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+
 static PyObject * widgetListboxAdd(snackWidget * s, PyObject * args) {
     char * text;
     
@@ -648,6 +956,18 @@ static PyObject * widgetListboxGet(snackWidget * s, PyObject * args) {
     return PyInt_FromLong((long) newtListboxGetCurrent(s->co));
 }
 
+static PyObject * widgetListboxSet(snackWidget * s, PyObject * args) {
+    int index;
+    
+    if (!PyArg_ParseTuple(args, "i", &index))
+	return NULL;
+
+    newtListboxSetCurrentByKey(s->co, (void *) index);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * widgetListboxSetW(snackWidget * s, PyObject * args) {
     int width;
 
@@ -663,6 +983,135 @@ static PyObject * widgetListboxSetW(snackWidget * s, PyObject * args) {
 static void emptyDestructor(PyObject * s) {
 }
 
+static snackWidget * checkboxTreeWidget(PyObject * s, PyObject * args) {
+    int height;
+    int scrollBar = 0;
+    snackWidget * widget;
+    
+    if (!PyArg_ParseTuple(args, "i|i", &height, &scrollBar))
+	return NULL;
+
+    widget = snackWidgetNew ();
+    widget->co = newtCheckboxTree(-1, -1, height,
+				  scrollBar ? NEWT_FLAG_SCROLL : 0);
+
+    widget->anint = 1;
+
+    return widget;
+}
+
+static PyObject * widgetCheckboxTreeAddItem(snackWidget * s, PyObject * args) {
+    char * text;
+    int selected = 0;
+    PyObject * pathList, * o;
+    int len;
+    int * path;
+    int i;
+
+    if (!PyArg_ParseTuple(args, "sOi", &text, &pathList, &selected))
+	return NULL;
+
+    len = PyTuple_Size(pathList);
+    path = alloca(sizeof(*path) * (len + 1));
+    for (i = 0; i < len; i++) {
+        o = PyTuple_GetItem(pathList, i);
+	path[i] = PyInt_AsLong(o);
+    }
+    path[len] = NEWT_ARG_LAST;
+
+    newtCheckboxTreeAddArray(s->co, text, (void *) s->anint,
+    			     selected ? NEWT_FLAG_SELECTED : 0, path);
+
+    return PyInt_FromLong(s->anint++);
+}
+
+static PyObject * widgetCheckboxTreeGetCur(snackWidget * s, PyObject * args) {
+    if (!PyArg_ParseTuple(args, ""))
+	return NULL;
+
+    return PyInt_FromLong((int) newtCheckboxTreeGetCurrent(s->co));
+}
+
+static PyObject * widgetCheckboxTreeSetEntry(snackWidget * s, PyObject * args) {
+    int data;
+    char *text;
+
+    if (!PyArg_ParseTuple(args, "is", &data, &text)) return NULL;
+
+    newtCheckboxTreeSetEntry(s->co, (void *)data, text);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * widgetCheckboxTreeSetEntryValue(snackWidget * s, PyObject * args) {
+    int data;
+    int isOn = 1;
+
+    if (!PyArg_ParseTuple(args, "i|i", &data, &isOn)) return NULL;
+
+    newtCheckboxTreeSetEntryValue(s->co, (void *)data,
+				  isOn ? NEWT_CHECKBOXTREE_SELECTED :
+					 NEWT_CHECKBOXTREE_UNSELECTED);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * widgetCheckboxTreeGetEntryValue(snackWidget * s, PyObject * args) {
+    int data;
+    int isOn = 0;
+    int isBranch = 0;
+    char selection;
+
+    if (!PyArg_ParseTuple(args, "i", &data)) return NULL;
+
+    selection = newtCheckboxTreeGetEntryValue(s->co, (void *)data);
+
+    if (selection == -1) return NULL;
+
+    switch (selection) {
+    case NEWT_CHECKBOXTREE_EXPANDED:
+	isOn = 1;
+    case NEWT_CHECKBOXTREE_COLLAPSED:
+	isBranch = 1;
+	break;
+    case NEWT_CHECKBOXTREE_UNSELECTED:
+	break;
+    default:
+	isOn = 1;
+	break;
+    }    
+    return Py_BuildValue("(ii)", isBranch, isOn);
+}
+
+static PyObject * widgetCheckboxTreeGetSel(snackWidget * s,
+					      PyObject * args) {
+    void ** selection;
+    int numselected;
+    int i;
+    PyObject * sel;
+
+    if (!PyArg_ParseTuple(args, ""))
+	return NULL;
+
+    selection = (void **) newtCheckboxTreeGetSelection(s->co, &numselected);
+
+    sel = PyList_New(0);
+    
+    if (!selection) {
+	return sel;
+    }
+
+    sel = PyList_New(0);
+    for (i = 0; i < numselected; i++) {
+    	PyList_Append(sel, PyInt_FromLong((int) selection[i]));
+    }
+    free(selection);
+
+    return sel;
+}
+
 void init_snack(void) {
     PyObject * d, * m;
 
@@ -679,6 +1128,7 @@ void init_snack(void) {
 
     PyDict_SetItemString(d, "FORM_EXIT_HOTKEY", PyString_FromString("hotkey"));
     PyDict_SetItemString(d, "FORM_EXIT_WIDGET", PyString_FromString("widget"));
+    PyDict_SetItemString(d, "FORM_EXIT_TIMER", PyString_FromString("timer"));
 
     PyDict_SetItemString(d, "KEY_F1", PyInt_FromLong(NEWT_KEY_F1));
     PyDict_SetItemString(d, "KEY_F2", PyInt_FromLong(NEWT_KEY_F2));
@@ -692,4 +1142,9 @@ void init_snack(void) {
     PyDict_SetItemString(d, "KEY_F10", PyInt_FromLong(NEWT_KEY_F10));
     PyDict_SetItemString(d, "KEY_F11", PyInt_FromLong(NEWT_KEY_F11));
     PyDict_SetItemString(d, "KEY_F12", PyInt_FromLong(NEWT_KEY_F12));
+
+    PyDict_SetItemString(d, "FLAG_DISABLED", PyInt_FromLong(NEWT_FLAG_DISABLED));
+    PyDict_SetItemString(d, "FLAGS_SET", PyInt_FromLong(NEWT_FLAGS_SET));
+    PyDict_SetItemString(d, "FLAGS_RESET", PyInt_FromLong(NEWT_FLAGS_RESET));
+    PyDict_SetItemString(d, "FLAGS_TOGGLE", PyInt_FromLong(NEWT_FLAGS_TOGGLE));
 }
